@@ -7,6 +7,8 @@
  * Author(s): Volker Sameske (sameske@de.ibm.com)
  */
 
+#define _GNU_SOURCE
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <getopt.h>
@@ -15,6 +17,7 @@
 #include <unistd.h>
 #include <errno.h>
 #include <dirent.h>
+#include <ftw.h>
 #include <sys/utsname.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -34,6 +37,10 @@
 static const char tool_name[] = "dasdinfo: zSeries DASD information program";
 static const char copyright_notice[] = "Copyright IBM Corp. 2007";
 
+/* needed because ftw can not pass arbitrary arguments */
+static char *searchbusid;
+static char *busiddir;
+
 struct volume_label {
 	char volkey[4];
 	char vollbl[4];
@@ -101,8 +108,8 @@ struct dasd_data
 
 static void dinfo_print_version (void)
 {
-	printf ("%s version %s.\n", tool_name, RELEASE_STRING);
-	printf ("%s.\n", copyright_notice);
+	printf ("%s version %s\n", tool_name, RELEASE_STRING);
+	printf ("%s\n", copyright_notice);
 }
 
 static void dinfo_print_usage(char *cmd)
@@ -110,15 +117,17 @@ static void dinfo_print_usage(char *cmd)
 	printf(""
 	       "Display specific information about a specified DASD device.\n"
 	       "\n"
-	       "Usage: %s [-a] [-u] [-l] [-e]\n"
+	       "Usage: %s [-a] [-u] [-x] [-l] [-e]\n"
 	       "                {-i <busid> | -b <blockdev> | -d <devnode>}\n"
 	       "       %s [-h] [-v]\n"
 	       "\n"
 	       "where:\n"
 	       "    -a|--all\n"
-	       "             same as -u -l\n"
+	       "             same as -u -x -l\n"
 	       "    -u|--uid\n"
-	       "             print DASD uid\n"
+	       "             print DASD uid (without z/VM minidisk token)\n"
+	       "    -x|--extended-uid\n"
+	       "             print DASD uid (including z/VM minidisk token)\n"
 	       "    -l|--label\n"
 	       "             print DASD volume label (volser)\n"
 	       "    -i|--busid <busid>\n"
@@ -250,6 +259,7 @@ static int dinfo_read_dasd_vlabel (char *device, struct volume_label *vlabel,
 	unsigned long vlabel_start;
 	char vollbl[5];
 	int f;
+	char *space;
 
 	if ((f = open(device, O_RDONLY)) < 0) {
 		printf("Could not open device node.\n");
@@ -294,12 +304,16 @@ static int dinfo_read_dasd_vlabel (char *device, struct volume_label *vlabel,
 	strncpy(vollbl, vlabel->vollbl, 4);
 	dinfo_ebcdic_dec(vollbl, vollbl, 4);
 
-	if ((strncmp(vollbl, "VOL1", 4) == 0) || (strncmp(vollbl, "LNX1", 4)
-	    == 0)) {
+	if ((strncmp(vollbl, "VOL1", 4) == 0) ||
+	    (strncmp(vollbl, "LNX1", 4) == 0) ||
+	    (strncmp(vollbl, "CMS1", 4) == 0)) {
 		strncpy(readbuf, vlabel->volid, 6);
 		dinfo_ebcdic_dec(readbuf, readbuf, 6);
+		space = strchr(readbuf,' ');
+		if (space)
+			*space = 0;
 	} else
-		strncpy(readbuf, "      ", 6);
+		strcpy(readbuf, "");
 
 
 	return 0;
@@ -446,32 +460,110 @@ static int dinfo_get_dev_from_blockdev(char *blockdev, dev_t *dev)
 	return 0;
 }
 
-static char *dinfo_get_blockdev_from_busid(char *busid, char **blkdev)
+static int
+dinfo_is_busiddir(const char *fpath, const struct stat *sb,
+		  int tflag, struct FTW *ftwbuf)
+{
+	char *tempdir;
+	char linkdir[128];
+	ssize_t i;
+	if (tflag != FTW_D || (strncmp((fpath + ftwbuf->base), searchbusid,
+				       strlen(searchbusid)) != 0))
+		return FTW_CONTINUE;
+	/*
+	 * ensure that the found entry is a busid and not a
+	 * subchannel ID
+	 * for large systems subchannel IDs may look like busids
+	 */
+	if (asprintf(&tempdir, "%s/driver", fpath) < 0)
+		return -1;
+	i = readlink(tempdir, linkdir, 128);
+	free(tempdir);
+	if ((i < 0) || (i >= 128))
+		return -1;
+	/* append '\0' because readlink returns non zero terminated string */
+	tempdir[i+1] = '\0';
+	if (strstr(linkdir, "dasd") == NULL)
+		return FTW_CONTINUE;
+	free(busiddir);
+	busiddir = strdup(fpath);
+	if (busiddir == NULL)
+		return -1;
+	return FTW_STOP;
+}
+
+static int
+dinfo_find_entry(const char *dir, const char *searchstring,
+		 char type, char **result)
 {
-	char tmp[READCHUNK];
 	DIR *directory = NULL;
 	struct dirent *dir_entry = NULL;
-
-	sprintf(tmp,"/sys/bus/ccw/devices/%s/", busid);
-	directory = opendir(tmp);
-
-	if (!directory)
-		return NULL;
-
+	directory = opendir(dir);
+	if (directory == NULL)
+		return -1;
 	while ((dir_entry = readdir(directory)) != NULL) {
-		if (strncmp(dir_entry->d_name, "block:", 6) == 0) {
-			if ((*blkdev = dinfo_malloc(_D_ALLOC_NAMLEN(dir_entry)))
-			    == NULL)
-				return NULL;
-			strcpy(*blkdev, strchr(dir_entry->d_name, ':') +
-			       sizeof(char));
-			break;
+		/* compare if the found entry has exactly the same name
+		   and type as searched */
+		if ((strncmp(dir_entry->d_name, searchstring,
+			     strlen(searchstring)) == 0)
+		    && (dir_entry->d_type & type)) {
+			*result = strdup(dir_entry->d_name);
+			if (*result == NULL)
+				goto out;
+			closedir(directory);
+			return 0; /* found */
 		}
 	}
-
+out:
 	closedir(directory);
+	return -1; /* nothing found or error */
+}
 
-	return *blkdev;
+static int
+dinfo_get_blockdev_from_busid(char *busid, char **blkdev)
+{
+	int flags = FTW_PHYS; /* do not follow links */
+	int rc = -1;
+
+	char *tempdir = NULL;
+	char *result = NULL;
+	char *sysfsdir = "/sys/devices/";
+
+	/* dinfo_is_devnode needs to know the busid */
+	searchbusid = busid;
+	if (nftw(sysfsdir, dinfo_is_busiddir, 200, flags) != FTW_STOP)
+		goto out;
+
+	/*
+	 * new sysfs: busid directory  contains a directory 'block'
+	 * which contains a directory 'dasdXXX'
+	 */
+	rc = dinfo_find_entry(busiddir, "block", DT_DIR, &result);
+	if (rc == 0) {
+		if (asprintf(&tempdir, "%s/%s/", busiddir, result) < 0) {
+			rc = -1;
+			goto out2;
+		}
+		rc = dinfo_find_entry(tempdir, "dasd", DT_DIR, blkdev);
+	} else {
+		/*
+		 * old sysfs: entry for busiddir contain a link
+		 * 'block:dasdXXX'
+		 */
+		rc = dinfo_find_entry(busiddir, "block:", DT_LNK, &result);
+		if (rc != 0)
+			goto out2;
+		*blkdev = strdup(strchr(result, ':') + 1);
+		if (*blkdev == NULL)
+			rc = -1;
+	}
+
+out:
+	free(tempdir);
+out2:
+	free(busiddir);
+	free(result);
+	return rc;
 }
 
 static int dinfo_get_uid_from_devnode(char **uidfile, char *devnode)
@@ -546,28 +638,32 @@ int main(int argc, char * argv[])
 	int export = 0;
 	int c;
 	int print_uid = 0;
+	int print_extended_uid = 0;
 	int print_vlabel = 0;
 	char *blockdev = NULL;
 	char *busid = NULL;
 	char *devnode = NULL;
 	struct volume_label vlabel;
+	char *srchuid;
+	int i;
 
 	while (1) {
 		int option_index = 0;
 		static struct option long_options[] = {
-			{"all",     0, 0, 'a'},
-			{"uid",     0, 0, 'u'},
-			{"label",   0, 0, 'l'},
-			{"busid",   1, 0, 'i'},
-			{"block",   1, 0, 'b'},
-			{"devnode", 1, 0, 'd'},
-			{"export",  0, 0, 'e'},
-			{"help",    0, 0, 'h'},
-			{"version", 0, 0, 'v'},
+			{"all",          0, 0, 'a'},
+			{"uid",          0, 0, 'u'},
+			{"extended-uid", 0, 0, 'x'},
+			{"label",        0, 0, 'l'},
+			{"busid",        1, 0, 'i'},
+			{"block",        1, 0, 'b'},
+			{"devnode",      1, 0, 'd'},
+			{"export",       0, 0, 'e'},
+			{"help",         0, 0, 'h'},
+			{"version",      0, 0, 'v'},
 			{0, 0, 0, 0}
 		};
 
-		c = getopt_long (argc, argv, "vhaeulb:i:d:",
+		c = getopt_long (argc, argv, "vhaeuxlb:i:d:",
 				 long_options, &option_index);
 		if (c == -1)
 			break;
@@ -576,10 +672,14 @@ int main(int argc, char * argv[])
 		case 'a':
 			print_uid = 1;
 			print_vlabel = 1;
+			print_extended_uid = 1;
 			break;
 		case 'u':
 			print_uid = 1;
 			break;
+		case 'x':
+			print_extended_uid = 1;
+			break;
 		case 'l':
 			print_vlabel = 1;
 			break;
@@ -629,14 +729,13 @@ int main(int argc, char * argv[])
 		return -1;
 	}
 
-	if (!print_uid && !print_vlabel) {
+	if (!print_uid && !print_extended_uid && !print_vlabel) {
 		printf("Error: no action specified (e.g. -u)\n");
 		return -1;
 	}
 
 	if (((readbuf = dinfo_malloc(readbuflen)) == NULL) ||
-	    ((uidfile = dinfo_malloc(readbuflen)) == NULL) ||
-	    ((device = dinfo_malloc(readbuflen)) == NULL))
+	    ((uidfile = dinfo_malloc(readbuflen)) == NULL))
 		return -1;
 
 	/* try to read the uid attribute */
@@ -656,10 +755,32 @@ int main(int argc, char * argv[])
 
 	if (print_uid) {
 		if (dinfo_read_dasd_uid(uidfile, readbuf, readbuflen) == 0) {
+			/* look for the 4th '.' and cut there */
+			srchuid = readbuf - 1;
+			for (i = 0; i < 4; ++i) {
+				srchuid = index(srchuid + 1, '.');
+				if (!srchuid)
+					break;
+			}
+			if (srchuid) {
+				srchuid[0] = '\n';
+				srchuid[1] = 0;
+			}
 			if (export) {
 				printf("ID_UID=%s",readbuf);
 			} else
 				printf("%s",readbuf);
+			if (!print_vlabel && !print_extended_uid)
+				goto out;
+		}
+	}
+
+	if (print_extended_uid) {
+		if (dinfo_read_dasd_uid(uidfile, readbuf, readbuflen) == 0) {
+			if (export) {
+				printf("ID_XUID=%s",readbuf);
+			} else
+				printf("%s",readbuf);
 			if (!print_vlabel)
 				goto out;
 		}
@@ -669,7 +790,7 @@ int main(int argc, char * argv[])
 	if (busid) {
 		char *blockdev_name = NULL;
 
-		if (dinfo_get_blockdev_from_busid(busid, &blockdev_name) == NULL)
+		if (dinfo_get_blockdev_from_busid(busid, &blockdev_name) != 0)
 			goto error;
 
 		if (dinfo_get_dev_from_blockdev(blockdev_name, &dev) != 0)
@@ -688,18 +809,16 @@ int main(int argc, char * argv[])
 			goto error;
 
 	} else if (devnode) {
+		if ((device = dinfo_malloc(readbuflen)) == NULL)
+		    return -1;
 		strcpy(device, devnode);
 	}
 
 	if (dinfo_read_dasd_vlabel(device, &vlabel, readbuf) == 0) {
 		if (export) {
-			printf("ID_SERIAL=%6.6s\n",readbuf);
+			printf("ID_SERIAL=%s\n",readbuf);
 		} else
-			printf("%6.6s\n", readbuf);
-
-		if (busid || blockdev)
-			dinfo_free_devnode(device);
-
+			printf("%s\n", readbuf);
 		goto out;
 	}
 
@@ -707,6 +826,9 @@ error:
 	printf("Error: could not read unique DASD ID\n");
 
 out:
+	if (device && (busid || blockdev))
+		dinfo_free_devnode(device);
+
 	free(uidfile);
 	free(device);
 	free(readbuf);
