@@ -41,7 +41,7 @@ struct name_lens_t {
 static int num_cpus;
 static int curr_small_max, curr_big_max;
 static int prev_small_max, prev_big_max;
-static int pg_to_kb_shift, sort_tbl_size;
+static unsigned int pg_to_kb_shift, sort_tbl_size;
 static float e_time;
 static struct timeval prev_time, curr_time;
 static struct cpudata_t cpudata;
@@ -62,7 +62,7 @@ static const char *pid_file = "/var/run/mon_procd.pid";
 /*
  * Clean up when SIGTERM or SIGINT received
  */
-void stop_procd(int a)
+void stop_procd(int UNUSED(a))
 {
 	remove(pid_file);
 	exit(0);
@@ -201,15 +201,11 @@ static int read_file(char *fname, char *buf, int size)
 	int num, fp;
 
 	fp = open(fname, O_RDONLY);
-	if (!fp) {
-		syslog(LOG_ERR, "failed %s open: %s\n", fname,
-			strerror(errno));
+	if (!fp)
 		return -1;
-	}
 
 	num = read(fp, buf, size);
 	if (num < 0) {
-		syslog(LOG_ERR, "failed %s read \n", fname);
 		close(fp);
 		return -1;
 	}
@@ -509,7 +505,7 @@ static int read_status(struct task_t *task)
 */
 static void cal_task_pcpu(struct task_t *task, const unsigned long long tics)
 {
-	int i, size;
+	unsigned int i, size;
 	__u64 etics;
 
 	if (proc_sum.task.total + 1 >= sort_tbl_size) {
@@ -798,7 +794,7 @@ static void task_usage(struct task_t *task)
 static void read_tasks(void)
 {
 	int size;
-	int i = 0, j = 0;
+	unsigned int i = 0, j = 0;
 	DIR *proc_dir;
 	struct direct *entry;
 	struct task_t task;
@@ -812,7 +808,7 @@ static void read_tasks(void)
 	while ((entry = readdir(proc_dir))) {
 		if (!entry->d_name)
 			break;
-		if (!isdigit(entry->d_name[0]))
+		if (!isdigit(entry->d_name[0]) || !atoi(entry->d_name))
 			continue;
 		memset(&task, 0, sizeof(struct task_t));
 		task.pid = atoi(entry->d_name);
@@ -956,9 +952,8 @@ static int parse_options(int argc, char **argv)
 			printf("%s", help_text);
 			exit(0);
 		case 'v':
-			printf( "mon_procd: version %s (%s)\n",
-				RELEASE_STRING, PROCD_VERSION);
-			printf("Copyright IBM Corp. 2007.\n");
+			printf( "mon_procd: version %s\n", RELEASE_STRING);
+			printf("Copyright IBM Corp. 2007\n");
 			exit(0);
 		case 'a':
 			attach = 1;
