@@ -47,17 +47,19 @@ void
 ProgressBar::displayProgress(uint64_t value, uint64_t maxValue)
 {
 	int j;
-	printf("%6lld of %6lld |", (long long)value, (long long)maxValue);
-	if(progressPercentage != (int)(value*100/maxValue))
-	{
-		/* percent value has changed */
-		progressPercentage = (value*100/maxValue);
-		for (j=1; j<=progressPercentage/2; j++)
-			printf("#");
-		for (j=progressPercentage/2+1; j<=50; j++)
-			printf("-");
-		printf("| %3d%%  ", progressPercentage);
+	char progress_bar[50];
+
+	if (progressPercentage == (int) (value * 100 / maxValue))
+		fprintf(stderr, "%6lld of %6lld |\r",
+			(long long) value, (long long) maxValue);
+	else {  /* percent value has changed */
+		progressPercentage = (value * 100 / maxValue);
+		for (j = 0; j < progressPercentage / 2; j++)
+			progress_bar[j] = '#';
+		for (j = progressPercentage / 2; j < 50; j++)
+			progress_bar[j] = '-';
+		fprintf(stderr, "%6lld of %6lld |%s| %3d%%  \r",
+			(long long) value, (long long) maxValue,
+			progress_bar, progressPercentage);
 	}
-	printf("\r");
-	fflush(stdout);
 }
