@@ -11,6 +11,7 @@
 #define LKCD_DUMP_H
 
 #include "dump.h"
+#include "zt_common.h"
 #include "register_content.h"
 
 
@@ -40,16 +41,15 @@
 
 #define GZIP_NOT_COMPRESSED -1
 
-
 class LKCDDump : public Dump
 {
 public:
 	LKCDDump(Dump*,const char*);
 	virtual ~LKCDDump(void){}
-	inline virtual void readMem(char* buf, int size) {
+	inline virtual void readMem(char* UNUSED(buf), int UNUSED(size)) {
 		throw(DumpException("LKCDDump::readMem() not implemented!"));
 	}
-	inline int seekMem(uint64_t offset){
+	inline int seekMem(uint64_t UNUSED(offset)){
 		throw(DumpException("LKCDDump::seekMem() not implemented!"));
 	}
 	inline virtual uint64_t getMemSize() const
