@@ -2,7 +2,7 @@
  * s390-tools/zipl/include/disk.h
  *   Functions to handle disk layout specific operations.
  *
- * Copyright IBM Corp. 2001, 2006.
+ * Copyright IBM Corp. 2001, 2009.
  *
  * Author(s): Carsten Otte <cotte@de.ibm.com>
  *            Peter Oberparleiter <Peter.Oberparleiter@de.ibm.com>
@@ -48,8 +48,7 @@ typedef enum {
 	disk_type_fba,
 	disk_type_diag,
 	disk_type_eckd_classic,
-	disk_type_eckd_compatible,
-	disk_type_unknown
+	disk_type_eckd_compatible
 } disk_type_t;
 
 /* from linux/hdregs.h */
@@ -60,6 +59,13 @@ struct hd_geometry {
 	unsigned long start;
 };
 
+/* Disk information source */
+typedef enum {
+	source_auto,
+	source_user,
+	source_script
+} source_t;
+
 /* Disk information type */
 struct disk_info {
 	disk_type_t type;
@@ -73,13 +79,19 @@ struct disk_info {
 	struct hd_geometry geo;
 	char* name;
 	char* drv_name;
+	source_t source;
 };
 
+struct job_target_data;
 
-int disk_get_info(const char* device, struct disk_info** info);
-int disk_get_info_from_file(const char* filename, struct disk_info** info);
+int disk_get_info(const char* device, struct job_target_data* target,
+		  struct disk_info** info);
+int disk_get_info_from_file(const char* filename,
+			    struct job_target_data* target,
+			    struct disk_info** info);
 void disk_free_info(struct disk_info* info);
 char* disk_get_type_name(disk_type_t type);
+int disk_is_large_volume(struct disk_info* info);
 int disk_cyl_from_blocknum(blocknum_t blocknum, struct disk_info* info);
 int disk_head_from_blocknum(blocknum_t blocknum, struct disk_info* info);
 int disk_sec_from_blocknum(blocknum_t blocknum, struct disk_info* info);
@@ -99,5 +111,6 @@ blocknum_t disk_compact_blocklist(disk_blockptr_t* list, blocknum_t count,
 blocknum_t disk_get_blocklist_from_file(const char* filename,
 					disk_blockptr_t** blocklist,
 					struct disk_info* pinfo);
+int disk_check_subchannel_set(int devno, dev_t device, char* dev_name);
 
 #endif /* not DISK_H */
