@@ -29,6 +29,8 @@ char* misc_strdup(const char* s);
 int misc_read(int fd, void* buffer, size_t count);
 int misc_read_file(const char* filename, char** buffer, size_t* size,
 		   int nil_terminate);
+int misc_read_special_file(const char* filename, char** buffer, size_t* size,
+			   int nil_terminate);
 int misc_write(int fd, const void* data, size_t count);
 int misc_get_file_buffer(const char* filename,
 			 struct misc_file_buffer* buffer);
