@@ -2,7 +2,7 @@
  * s390-tools/zipl/src/bootmap.c
  *   Functions to build the bootmap file.
  *
- * Copyright IBM Corp. 2001, 2006.
+ * Copyright IBM Corp. 2001, 2009.
  *
  * Author(s): Carsten Otte <cotte@de.ibm.com>
  *            Peter Oberparleiter <Peter.Oberparleiter@de.ibm.com>
@@ -71,7 +71,6 @@ get_blockptr_size(struct disk_info* info)
 	case disk_type_eckd_compatible:
 		return sizeof(struct eckd_blockptr);
 	case disk_type_diag:
-	case disk_type_unknown:
 		break;
 	}
 	return 0;
@@ -105,13 +104,13 @@ bootmap_store_blockptr(void* buffer, disk_blockptr_t* ptr,
 		case disk_type_eckd_compatible:
 			eckd = (struct eckd_blockptr *) buffer;
 			eckd->cyl = ptr->chs.cyl;
-			eckd->head = ptr->chs.head;
+			eckd->head = ptr->chs.head |
+				     ((ptr->chs.cyl >> 12) & 0xfff0);
 			eckd->sec = ptr->chs.sec;
 			eckd->size = ptr->chs.size;
 			eckd->blockct = ptr->chs.blockct;
 			break;
 		case disk_type_diag:
-		case disk_type_unknown:
 			break;
 		}
 	}
@@ -299,7 +298,8 @@ struct component_loc {
 static int
 add_component_file(int fd, const char* filename, address_t load_address,
 		   off_t offset, void* component, int add_files,
-		   struct disk_info* info, struct component_loc *location)
+		   struct disk_info* info, struct job_target_data* target,
+		   struct component_loc *location)
 {
 	struct disk_info* file_info;
 	struct component_loc loc;
@@ -337,7 +337,7 @@ add_component_file(int fd, const char* filename, address_t load_address,
 		}
 	} else {
 		/* Make sure file is on correct device */
-		rc = disk_get_info_from_file(filename, &file_info);  
+		rc = disk_get_info_from_file(filename, target, &file_info);
 		if (rc)
 			return -1;
 		if (file_info->device != info->device) {
@@ -473,7 +473,7 @@ check_component_overlap(const char *name[], struct component_loc *loc, int num)
 static int
 add_ipl_program(int fd, struct job_ipl_data* ipl, disk_blockptr_t* program,
 		int verbose, int add_files, component_header_type type,
-		struct disk_info* info)
+		struct disk_info* info, struct job_target_data* target)
 {
 	struct stat stats;
 	void* table;
@@ -501,7 +501,7 @@ add_ipl_program(int fd, struct job_ipl_data* ipl, disk_blockptr_t* program,
 	}
 	rc = add_component_file(fd, ipl->image, ipl->image_addr,
 				KERNEL_HEADER_SIZE, VOID_ADD(table, offset),
-				add_files, info, &comp_loc[0]);
+				add_files, info, target, &comp_loc[0]);
 	if (rc) {
 		error_text("Could not add image file '%s'", ipl->image);
 		free(table);
@@ -543,7 +543,7 @@ add_ipl_program(int fd, struct job_ipl_data* ipl, disk_blockptr_t* program,
 		rc = add_component_file(fd, ipl->ramdisk,
 					ipl->ramdisk_addr, 0,
 					VOID_ADD(table, offset),
-					add_files, info, &comp_loc[2]);
+					add_files, info, target, &comp_loc[2]);
 		if (rc) {
 			error_text("Could not add ramdisk '%s'",
 				   ipl->ramdisk);
@@ -595,7 +595,8 @@ add_ipl_program(int fd, struct job_ipl_data* ipl, disk_blockptr_t* program,
 static int
 add_segment_program(int fd, struct job_segment_data* segment,
 		    disk_blockptr_t* program, int verbose, int add_files,
-		    component_header_type type, struct disk_info* info)
+		    component_header_type type, struct disk_info* info,
+		    struct job_target_data* target)
 {
 	const char *comp_name[1] = {"segment file"};
 	struct component_loc comp_loc[1];
@@ -619,7 +620,7 @@ add_segment_program(int fd, struct job_segment_data* segment,
 	}
 	rc = add_component_file(fd, segment->segment, segment->segment_addr, 0,
 				VOID_ADD(table, offset), add_files, info,
-				&comp_loc[0]);
+				target, &comp_loc[0]);
 	if (rc) {
 		error_text("Could not add segment file '%s'",
 			   segment->segment);
@@ -662,14 +663,15 @@ create_dump_fs_parmline(const char* parmline, const char* root_dev,
 
 static int
 get_dump_fs_parmline(char* partition, char* parameters, uint64_t mem,
-		     char** result, struct disk_info* target_info)
+		     struct disk_info* target_info,
+		     struct job_target_data* target, char** result)
 {
 	char* buffer;
 	struct disk_info* info;
 	int rc;
 
 	/* Get information about partition */
-	rc = disk_get_info(partition, &info);
+	rc = disk_get_info(partition, target, &info);
 	if (rc) {
 		error_text("Could not get information for dump partition '%s'",
 			   partition);
@@ -701,7 +703,7 @@ static int
 add_dump_fs_program(int fd, struct job_dump_fs_data* dump_fs,
 		    disk_blockptr_t* program, int verbose,
 		    int add_files, component_header_type type,
-		    struct disk_info* info)
+		    struct disk_info* info, struct job_target_data* target)
 {
 	struct job_ipl_data ipl;
 	int rc;
@@ -726,12 +728,12 @@ add_dump_fs_program(int fd, struct job_dump_fs_data* dump_fs,
 
 	/* Get file system dump parmline */
 	rc = get_dump_fs_parmline(dump_fs->partition, dump_fs->parmline,
-				  dump_fs->mem, &ipl.parmline, info);
+				  dump_fs->mem, info, target, &ipl.parmline);
 	if (rc)
 		return rc;
 	ipl.parm_addr = DEFAULT_PARMFILE_ADDRESS;
 	return add_ipl_program(fd, &ipl, program, verbose, 1,
-			       type, info);
+			       type, info, target);
 }
 
 
@@ -764,7 +766,7 @@ build_program_table(int fd, struct job_data* job, disk_blockptr_t* pointer,
 		rc = add_ipl_program(fd, &job->data.ipl, &table[0],
 				     verbose || job->command_line,
 				     job->add_files, component_header_ipl,
-				     info);
+				     info, &job->target);
 		break;
 	case job_segment:
 		if (job->command_line)
@@ -775,7 +777,7 @@ build_program_table(int fd, struct job_data* job, disk_blockptr_t* pointer,
 		rc = add_segment_program(fd, &job->data.segment, &table[0],
 					 verbose || job->command_line,
 					 job->add_files, component_header_ipl,
-					 info);
+					 info, &job->target);
 		break;
 	case job_dump_fs:
 		if (job->command_line)
@@ -786,7 +788,7 @@ build_program_table(int fd, struct job_data* job, disk_blockptr_t* pointer,
 		rc = add_dump_fs_program(fd, &job->data.dump_fs, &table[0],
 					 verbose || job->command_line,
 					 job->add_files, component_header_dump,
-					 info);
+					 info, &job->target);
 		break;
 	case job_menu:
 		printf("Building menu '%s'\n", job->name);
@@ -805,7 +807,7 @@ build_program_table(int fd, struct job_data* job, disk_blockptr_t* pointer,
 					&table[job->data.menu.entry[i].pos],
 					verbose || job->command_line,
 					job->add_files,	component_header_ipl,
-					info);
+					info, &job->target);
 				break;
 			case job_dump_fs:
 				printf("Adding #%d: fs-dump section '%s'%s\n",
@@ -819,12 +821,13 @@ build_program_table(int fd, struct job_data* job, disk_blockptr_t* pointer,
 					&table[job->data.menu.entry[i].pos],
 					verbose || job->command_line,
 					job->add_files, component_header_dump,
-					info);
+					info, &job->target);
 				break;
 			case job_print_usage:
 			case job_print_version:
 			case job_segment:
 			case job_dump_partition:
+			case job_mvdump:
 			case job_menu:
 			case job_ipl_tape:
 				rc = -1;
@@ -890,7 +893,8 @@ bootmap_create(struct job_data* job, disk_blockptr_t* program_table,
 	int rc;
 
 	/* Get full path of bootmap file */
-	filename = misc_make_path(job->bootmap_dir, BOOTMAP_TEMPLATE_FILENAME);
+	filename = misc_make_path(job->target.bootmap_dir,
+			BOOTMAP_TEMPLATE_FILENAME);
 	if (filename == NULL)
 		return -1;
 	/* Create temporary bootmap file */
@@ -904,49 +908,14 @@ bootmap_create(struct job_data* job, disk_blockptr_t* program_table,
 	/* Retrieve target device information. Note that we have to
 	 * call disk_get_info_from_file() to also get the file system
 	 * block size. */
-	rc = disk_get_info_from_file(filename, &info);
+	rc = disk_get_info_from_file(filename, &job->target, &info);
 	if (rc) {
 		close(fd);
 		free(filename);
 		return -1;
 	}
 	/* Check for supported disk and driver types */
-	switch (info->type) {
-	case disk_type_scsi:
-		if (info->drv_name && (strcmp(info->drv_name, "sd") != 0)) {
-			error_reason("Unsupported device driver '%s'",
-				     info->drv_name);
-			disk_free_info(info);
-			close(fd);
-			free(filename);
-			return -1;
-		}
-		break;
-	case disk_type_fba:
-	case disk_type_eckd_classic:
-	case disk_type_eckd_compatible:
-		if (info->drv_name && (strcmp(info->drv_name, "dasd") != 0)) {
-			error_reason("Unsupported device driver '%s'",
-				     info->drv_name);
-			disk_free_info(info);
-			close(fd);
-			free(filename);
-			return -1;
-		}
-		/* Check for valid CHS geometry data. */
-		if ((info->geo.cylinders == 0) || (info->geo.heads == 0) ||
-		    (info->geo.sectors == 0)) {
-			error_reason("Invalid disk geometry (CHS=%d/%d/%d)",
-				     info->geo.cylinders, info->geo.heads,
-				     info->geo.sectors);
-			disk_free_info(info);
-			close(fd);
-			free(filename);
-			return -1;
-		}
-		break;
-	case disk_type_diag:
-	case disk_type_unknown:
+	if ((info->source == source_auto) && (info->type == disk_type_diag)) {
 		error_reason("Unsupported disk type (%s)",
 			     disk_get_type_name(info->type));
 		disk_free_info(info);
@@ -976,7 +945,7 @@ bootmap_create(struct job_data* job, disk_blockptr_t* program_table,
 			return rc;
 		}
 	}
-	printf("Building bootmap in '%s'%s\n", job->bootmap_dir,
+	printf("Building bootmap in '%s'%s\n", job->target.bootmap_dir,
 	       job->add_files ? " (files will be added to bootmap file)" :
 	       "");
 	/* Write bootmap header */
@@ -1052,7 +1021,6 @@ bootmap_create(struct job_data* job, disk_blockptr_t* program_table,
 	case disk_type_scsi:
 	case disk_type_eckd_compatible:
 	case disk_type_diag:
-	case disk_type_unknown:
 		*stage2_list = NULL;
 		*stage2_count = 0;
 		break;
@@ -1064,7 +1032,8 @@ bootmap_create(struct job_data* job, disk_blockptr_t* program_table,
 				"file %s!\n", filename);
 	} else {
 		/* Rename to final bootmap name */
-		mapname = misc_make_path(job->bootmap_dir, BOOTMAP_FILENAME);
+		mapname = misc_make_path(job->target.bootmap_dir,
+				BOOTMAP_FILENAME);
 		if (mapname == NULL) {
 			misc_free_temp_dev(device);
 			disk_free_info(info);
