@@ -2,13 +2,14 @@
  * s390-tools/zipl/src/disk.c
  *   Functions to handle disk layout specific operations.
  *
- * Copyright IBM Corp. 2001, 2006.
+ * Copyright IBM Corp. 2001, 2009.
  *
  * Author(s): Carsten Otte <cotte@de.ibm.com>
  *            Peter Oberparleiter <Peter.Oberparleiter@de.ibm.com>
  */
 
 #include "disk.h"
+#include "job.h"
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -20,6 +21,8 @@
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <dirent.h>
+#include <limits.h>
 
 #include "error.h"
 #include "misc.h"
@@ -62,16 +65,54 @@ struct dasd_information {
 	char configuration_data[256];	/* from read_configuration_data */
 };
 
+static int
+disk_determine_dasd_type(struct disk_info *data,
+	struct dasd_information dasd_info)
+{
+	if (strncmp(dasd_info.type, "FBA ",4) == 0)
+		data->type = disk_type_fba;
+	else if (strncmp(dasd_info.type, "DIAG",4) == 0)
+		data->type = disk_type_diag;
+	else if (strncmp(dasd_info.type, "ECKD",4) == 0) {
+		if (dasd_info.FBA_layout)
+			data->type = disk_type_eckd_classic;
+		else
+			data->type = disk_type_eckd_compatible;
+	} else {
+		error_reason("Unknown DASD type");
+		return -1;
+	}
+	return 0;
+}
+
+/* Return non-zero for ECKD type. */
 int
-disk_get_info(const char* device, struct disk_info** info)
+disk_is_eckd(disk_type_t type)
+{
+	return (type == disk_type_eckd_classic ||
+		type == disk_type_eckd_compatible);
+}
+
+int
+disk_get_info(const char* device, struct job_target_data* target,
+	      struct disk_info** info)
 {
 	struct stat stats;
+	struct stat script_stats;
 	struct proc_part_entry part_entry;
 	struct proc_dev_entry dev_entry;
 	struct dasd_information dasd_info;
 	struct disk_info *data;
 	int fd;
 	long devsize;
+	FILE *fh;
+	char *script_pre = "/lib/s390-tools/zipl_helper.";
+	char script_file[80];
+	char ppn_cmd[80];
+	char buffer[80];
+	char value[40];
+	int majnum, minnum;
+	int checkparm;
 
 	/* Get file information */
 	if (stat(device, &stats)) {
@@ -91,80 +132,195 @@ disk_get_info(const char* device, struct disk_info** info)
 		return -1;
 	}
 	memset((void *) data, 0, sizeof(struct disk_info));
+	/* Try to get device driver name */
+	if (proc_dev_get_entry(stats.st_rdev, 1, &dev_entry) == 0) {
+		data->drv_name = misc_strdup(dev_entry.name);
+		proc_dev_free_entry(&dev_entry);
+	} else {
+		fprintf(stderr, "Warning: Could not determine driver name for "
+			"major %d from /proc/devices\n", major(stats.st_rdev));
+		fprintf(stderr, "Warning: Preparing a logical device for boot "
+			"might fail\n");
+	}
+	data->source = source_user;
+	/* Check if targetbase script is available */
+	strcpy(script_file, script_pre);
+	if (data->drv_name) {
+		strcat(script_file, data->drv_name);
+	}
+	if ((target->targetbase == NULL) &&
+	    (!stat(script_file, &script_stats))) {
+		data->source = source_script;
+		/* Run targetbase script */
+		strcpy(ppn_cmd, script_file);
+		strcat(ppn_cmd, " ");
+		strcat(ppn_cmd, target->bootmap_dir);
+		printf("Run %s\n", ppn_cmd);
+		fh = popen(ppn_cmd, "r");
+		if (fh == NULL) {
+			error_reason("Failed to run popen(%s,\"r\",)");
+			goto out_close;
+		}
+		checkparm = 0;
+		while (fgets(buffer, 80, fh) != NULL) {
+			if (sscanf(buffer, "targetbase=%s", value) == 1) {
+				target->targetbase = misc_strdup(value);
+				checkparm++;
+			}
+			if (sscanf(buffer, "targettype=%s", value) == 1) {
+				type_from_target(value, &target->targettype);
+				checkparm++;
+			}
+			if (sscanf(buffer, "targetgeometry=%s", value) == 1) {
+				target->targetcylinders =
+					atoi(strtok(value, ","));
+				target->targetheads = atoi(strtok(NULL, ","));
+				target->targetsectors = atoi(strtok(NULL, ","));
+				checkparm++;
+			}
+			if (sscanf(buffer, "targetblocksize=%s", value) == 1) {
+				target->targetblocksize = atoi(value);
+				checkparm++;
+			}
+			if (sscanf(buffer, "targetoffset=%s", value) == 1) {
+				target->targetoffset = atol(value);
+				checkparm++;
+			}
+		}
+		if (pclose(fh) == -1) {
+			error_reason("Failed to run pclose");
+			goto out_close;
+		}
+		if ((!disk_is_eckd(target->targettype) && checkparm < 4) ||
+		    (disk_is_eckd(target->targettype) && checkparm != 5)) {
+			error_reason("Target parameters missing from script");
+			goto out_close;
+		}
+	}
+
 	/* Get disk geometry. Note: geo.start contains a sector number
 	 * offset measured in physical blocks, not sectors (512 bytes) */
-	if (ioctl(fd, HDIO_GETGEO, &data->geo)) {
-		error_reason("Could not get disk geometry");
-		free(data);
-		close(fd);
-		return -1;
+	if (target->targetbase != NULL) {
+		data->geo.heads     = target->targetheads;
+		data->geo.sectors   = target->targetsectors;
+		data->geo.cylinders = target->targetcylinders;
+		data->geo.start     = target->targetoffset;
+	} else {
+		data->source = source_auto;
+		if (ioctl(fd, HDIO_GETGEO, &data->geo)) {
+			error_reason("Could not get disk geometry");
+			goto out_close;
+		}
 	}
-	/* Get DASD information */
-	if (ioctl(fd, BIODASDINFO, &dasd_info))
+	if ((data->source == source_user) || (data->source == source_script)) {
 		data->devno = -1;
-	else
-		data->devno = dasd_info.devno;
-
-	/* Determine disk type */
-	if (data->devno == -1)
-		/* FIXME: SCSI check? */
-		data->type = disk_type_scsi;
-	else if (strncmp(dasd_info.type, "FBA ",4) == 0)
-		data->type = disk_type_fba;
-	else if (strncmp(dasd_info.type, "DIAG",4) == 0)
-		data->type = disk_type_diag;
-	else if (strncmp(dasd_info.type, "ECKD",4) == 0) {
-		if (dasd_info.FBA_layout)
-			data->type = disk_type_eckd_classic;
+		data->phy_block_size = target->targetblocksize;
+	} else {
+		/* Get DASD information */
+		if (ioctl(fd, BIODASDINFO, &dasd_info))
+			data->devno = -1;
 		else
-			data->type = disk_type_eckd_compatible;
-	} else
-		data->type = disk_type_unknown;
-	/* Get physical block size */
-	if (ioctl(fd, BLKSSZGET, &data->phy_block_size)) {
-		error_reason("Could not get blocksize");
-		free(data);
-		close(fd);
-		return -1;
+			data->devno = dasd_info.devno;
+		/* Get physical block size */
+		if (ioctl(fd, BLKSSZGET, &data->phy_block_size)) {
+			error_reason("Could not get blocksize");
+			goto out_close;
+		}
 	}
 	/* Get size of device in sectors (512 byte) */
 	if (ioctl(fd, BLKGETSIZE, &devsize)) {
 		error_reason("Could not get device size");
-		close(fd);
-		free(data);
-		return -1;
+		goto out_close;
 	}
-	/* Convert device size to size in physical blocks */
-	data->phy_blocks = devsize / (data->phy_block_size / 512);
-	/* Set device and partition devno */
-	if (data->type == disk_type_scsi) {
+	if ((data->source == source_user) || (data->source == source_script)) {
+		data->type = target->targettype;
+		data->partnum = 0;
+		/* Get file information */
+		if (sscanf(target->targetbase, "%d:%d", &majnum, &minnum) == 2)
+			data->device = makedev(majnum, minnum);
+		else {
+			if (stat(target->targetbase, &stats)) {
+				error_reason(strerror(errno));
+				goto out_close;
+			}
+			if (!S_ISBLK(stats.st_mode)) {
+				error_reason("Target base device '%s' is not "
+					     "a block device",
+					     target->targetbase);
+				goto out_close;
+			}
+			data->device = stats.st_rdev;
+		}
+		goto type_determined;
+	}
+	/* Determine disk type */
+	if (!data->drv_name) {
+		/* Driver name cannot be read */
+		if (data->devno == -1) {
+			if (data->geo.start) {
+				/* SCSI partition */
+				data->type = disk_type_scsi;
+				data->partnum = stats.st_rdev & SCSI_PARTN_MASK;
+				data->device = stats.st_rdev & ~SCSI_PARTN_MASK;
+			} else {
+				/* SCSI disk */
+				data->type = disk_type_scsi;
+				data->partnum = 0;
+				data->device = stats.st_rdev;
+			}
+		} else {
+			/* DASD */
+			if (disk_determine_dasd_type(data, dasd_info))
+				goto out_close;
+			data->partnum = stats.st_rdev & DASD_PARTN_MASK;
+			data->device = stats.st_rdev & ~DASD_PARTN_MASK;
+		}
+	} else if (strcmp(data->drv_name, "dasd") == 0) {
+		/* Driver name is 'dasd' */
+		if (data->devno == -1) {
+			error_reason("Could not determine DASD type");
+			goto out_close;
+		}
+		if (disk_determine_dasd_type(data, dasd_info))
+			goto out_close;
+		data->partnum = stats.st_rdev & DASD_PARTN_MASK;
+		data->device = stats.st_rdev & ~DASD_PARTN_MASK;
+	} else if (strcmp(data->drv_name, "sd") == 0 ||
+		   strcmp(data->drv_name, "virtblk") == 0) {
+		/* Driver name is 'sd' or 'virtblk' */
+		if (data->devno != -1) {
+			error_reason("Unsupported device driver '%s' "
+				     "for disk type DASD", data->drv_name);
+			goto out_close;
+		}
+		data->type = disk_type_scsi;
 		data->partnum = stats.st_rdev & SCSI_PARTN_MASK;
 		data->device = stats.st_rdev & ~SCSI_PARTN_MASK;
 	} else {
-		data->partnum = stats.st_rdev & DASD_PARTN_MASK;
-		data->device = stats.st_rdev & ~DASD_PARTN_MASK;
+		/* Driver name is unknown */
+		error_reason("Unsupported device driver '%s'", data->drv_name);
+		goto out_close;
 	}
+
+type_determined:
+	/* Check for valid CHS geometry data. */
+	if (disk_is_eckd(data->type) && (data->geo.cylinders == 0 ||
+	    data->geo.heads == 0 || data->geo.sectors == 0)) {
+		error_reason("Invalid disk geometry (CHS=%d/%d/%d)",
+			     data->geo.cylinders, data->geo.heads,
+			     data->geo.sectors);
+		goto out_close;
+	}
+	/* Convert device size to size in physical blocks */
+	data->phy_blocks = devsize / (data->phy_block_size / 512);
 	if (data->partnum != 0)
 		data->partition = stats.st_rdev;
 	/* Try to get device name */
 	if (proc_part_get_entry(data->device, &part_entry) == 0) {
 		data->name = misc_strdup(part_entry.name);
 		proc_part_free_entry(&part_entry);
-		if (data->name == NULL) {
-			close(fd);
-			free(data);
-			return -1;
-		}
-	}
-	/* Try to get device driver name */
-	if (proc_dev_get_entry(data->device, 1, &dev_entry) == 0) {
-		data->drv_name = misc_strdup(dev_entry.name);
-		proc_dev_free_entry(&dev_entry);
-		if (data->name == NULL) {
-			close(fd);
-			free(data);
-			return -1;
-		}
+		if (data->name == NULL)
+			goto out_close;
 	}
 	/* There is no easy way to find out whether there is a file system
 	 * on this device, so we set the respective block size to an invalid
@@ -173,11 +329,17 @@ disk_get_info(const char* device, struct disk_info** info)
 	close(fd);
 	*info = data;
 	return 0;
+out_close:
+	close(fd);
+	free(data);
+	return -1;
+
 }
 
 
 int
-disk_get_info_from_file(const char* filename, struct disk_info** info)
+disk_get_info_from_file(const char* filename, struct job_target_data* target,
+			struct disk_info** info)
 {
 	struct stat stats;
 	char* device;
@@ -207,7 +369,7 @@ disk_get_info_from_file(const char* filename, struct disk_info** info)
 	if (rc)
 		return -1;
 	/* Get device info */
-	rc = disk_get_info(device, info);
+	rc = disk_get_info(device, target, info);
 	if (rc == 0)
 		(*info)->fs_block_size = blocksize;
 	/* Clean up */
@@ -342,9 +504,6 @@ disk_blockptr_from_blocknum(disk_blockptr_t* ptr, blocknum_t blocknum,
 		ptr->chs.size = info->phy_block_size;
 		ptr->chs.blockct = 0;
 		break;
-	case disk_type_unknown:
-		/* Should not happen */
-		break;
 	}
 }
 
@@ -468,10 +627,24 @@ disk_get_type_name(disk_type_t type)
 }
 
 
+/* Return non-zero for ECKD large volumes. */
+int
+disk_is_large_volume(struct disk_info* info)
+{
+	return (info->type == disk_type_eckd_classic ||
+	        info->type == disk_type_eckd_compatible) &&
+		info->geo.cylinders == 0xfffe;
+}
+
+
 /* Print textual representation of INFO contents. */
 void
 disk_print_info(struct disk_info* info)
 {
+	char footnote[4] = "";
+	if ((info->source == source_user) || (info->source == source_script))
+		strcpy(footnote, " *)");
+
 	printf("  Device..........................: ");
 	disk_print_devt(info->device);
 	printf("\n");
@@ -481,39 +654,50 @@ disk_print_info(struct disk_info* info)
 		printf("\n");
 	}
 	if (info->name) {
-		printf("  Device name.....................: %s\n",
-		       info->name);
+		printf("  Device name.....................: %s%s\n",
+		       info->name, footnote);
 	}
 	if (info->drv_name) {
 		printf("  Device driver name..............: %s\n",
 		       info->drv_name);
 	}
-	if ((info->type == disk_type_fba) ||
-	    (info->type == disk_type_diag) ||
-	    (info->type == disk_type_eckd_classic) ||
-	    (info->type == disk_type_eckd_compatible)) {
+	if (((info->type == disk_type_fba) ||
+	     (info->type == disk_type_diag) ||
+	     (info->type == disk_type_eckd_classic) ||
+	     (info->type == disk_type_eckd_compatible)) &&
+	     (info->source == source_auto)) {
 		printf("  DASD device number..............: %04x\n",
 		       info->devno);
 	}
 	printf("  Type............................: disk %s\n",
 	       (info->partnum != 0) ? "partition" : "device");
-	printf("  Disk layout.....................: %s\n",
-	       disk_get_type_name(info->type));
-	printf("  Geometry - heads................: %d\n",
-	       info->geo.heads);
-	printf("  Geometry - sectors..............: %d\n",
-	       info->geo.sectors);
-	printf("  Geometry - cylinders............: %d\n",
-	       info->geo.cylinders);
-	printf("  Geometry - start................: %ld\n",
-	       info->geo.start);
+	printf("  Disk layout.....................: %s%s\n",
+	       disk_get_type_name(info->type), footnote);
+	printf("  Geometry - heads................: %d%s\n",
+	       info->geo.heads, footnote);
+	printf("  Geometry - sectors..............: %d%s\n",
+	       info->geo.sectors, footnote);
+	if (disk_is_large_volume(info)) {
+		/* ECKD large volume. There is not enough information
+		 * available in INFO to calculate disk cylinder size. */
+		printf("  Geometry - cylinders............: > 65534\n");
+	} else {
+		printf("  Geometry - cylinders............: %d%s\n",
+		       info->geo.cylinders, footnote);
+	}
+	printf("  Geometry - start................: %ld%s\n",
+	       info->geo.start, footnote);
 	if (info->fs_block_size >= 0)
 		printf("  File system block size..........: %d\n",
 		       info->fs_block_size);
-	printf("  Physical block size.............: %d\n",
-	       info->phy_block_size);
+	printf("  Physical block size.............: %d%s\n",
+	       info->phy_block_size, footnote);
 	printf("  Device size in physical blocks..: %ld\n",
 	       (long) info->phy_blocks);
+	if (info->source == source_user)
+		printf("  *) Data provided by user.\n");
+	if (info->source == source_script)
+		printf("  *) Data provided by script.\n");
 }
 
 
@@ -581,7 +765,6 @@ can_merge_blocks(disk_blockptr_t* first, disk_blockptr_t* second,
 			return 1;
 		break;
 	case disk_type_diag:
-	case disk_type_unknown:
 		break;
 	}
 	return 0;
@@ -605,7 +788,6 @@ merge_blocks(disk_blockptr_t* first, disk_blockptr_t* second,
 		first->chs.blockct += second->chs.blockct + 1;
 		break;
 	case disk_type_diag:
-	case disk_type_unknown:
 		/* Should not happen */
 		break;
 	}
@@ -691,3 +873,62 @@ disk_get_blocklist_from_file(const char* filename, disk_blockptr_t** blocklist,
 	*blocklist = list;
 	return count;
 }
+
+/* Check whether input device is in subchannel set 0.
+ * Path to "dev" attribute containing the major/minor number depends on
+ * whether option CONFIG_SYSFS_DEPRECATED is set or not */
+
+int disk_check_subchannel_set(int devno, dev_t device, char* dev_name)
+{
+	struct dirent *direntp;
+	DIR* fdd;
+	static const char sys_bus_ccw_dev_filename[] = "/sys/bus/ccw/devices";
+	char dev_file[PATH_MAX];
+	char *buffer;
+	int minor, major;
+
+	snprintf(dev_file, PATH_MAX, "%s/0.0.%04x", sys_bus_ccw_dev_filename,
+		 devno);
+	fdd = opendir(dev_file);
+	if (!fdd)
+		goto out_with_warning;
+	while ((direntp = readdir(fdd)))
+		if (strncmp(direntp->d_name, "block:", 6) == 0)
+			break;
+	if (direntp != NULL)
+		snprintf(dev_file, PATH_MAX, "%s/0.0.%04x/%s/dev",
+			 sys_bus_ccw_dev_filename, devno, direntp->d_name);
+	else {
+		closedir(fdd);
+		snprintf(dev_file, PATH_MAX, "%s/0.0.%04x/block",
+			 sys_bus_ccw_dev_filename, devno);
+		fdd = opendir(dev_file);
+		if (!fdd)
+			goto out_with_warning;
+		while ((direntp = readdir(fdd)))
+			if (strncmp(direntp->d_name, "dasd", 4) == 0)
+				break;
+		if (direntp == NULL)
+			goto out_with_warning;
+		snprintf(dev_file, PATH_MAX, "%s/0.0.%04x/block/%s/dev",
+			 sys_bus_ccw_dev_filename, devno, direntp->d_name);
+	}
+	closedir(fdd);
+	if (misc_read_special_file(dev_file, &buffer, NULL, 1))
+		goto out_with_warning;
+	if (sscanf(buffer, "%i:%i", &major, &minor) != 2) {
+		free(buffer);
+		goto out_with_warning;
+	}
+	free(buffer);
+	if (makedev(major, minor) != device) {
+		error_reason("Dump target '%s' must belong to "
+			     "subchannel set 0.", dev_name);
+		return -1;
+	}
+	return 0;
+out_with_warning:
+	fprintf(stderr, "Warning: Could not determine whether dump target %s "
+		"belongs to subchannel set 0.\n", dev_name);
+	return 0;
+}
