@@ -3,7 +3,7 @@
  *   Functions and data structures representing the actual 'job' that the
  *   user wants us to execute.
  *
- * Copyright IBM Corp. 2001, 2006.
+ * Copyright IBM Corp. 2001, 2009.
  *
  * Author(s): Carsten Otte <cotte@de.ibm.com>
  *            Peter Oberparleiter <Peter.Oberparleiter@de.ibm.com>
@@ -29,12 +29,18 @@
 static struct option options[] = {
 	{ "config",		required_argument,	NULL, 'c'},
 	{ "target",		required_argument,	NULL, 't'},
+	{ "targetbase",         required_argument,      NULL, 0xaa},
+	{ "targettype",         required_argument,      NULL, 0xab},
+	{ "targetgeometry",     required_argument,      NULL, 0xac},
+	{ "targetblocksize",    required_argument,      NULL, 0xad},
+	{ "targetoffset",       required_argument,      NULL, 0xae},
 	{ "image", 		required_argument,	NULL, 'i'},
 	{ "ramdisk",		required_argument,	NULL, 'r'},
 	{ "parmfile",		required_argument,	NULL, 'p'},
 	{ "parameters",		required_argument,	NULL, 'P'},
 	{ "dumpto",		required_argument,	NULL, 'd'},
 	{ "dumptofs",		required_argument,	NULL, 'D'},
+	{ "mvdump",		required_argument,	NULL, 'M'},
 	{ "segment",		required_argument,	NULL, 's'},
 	{ "menu",		required_argument,	NULL, 'm'},
 	{ "help",		no_argument,		NULL, 'h'},
@@ -44,11 +50,12 @@ static struct option options[] = {
 	{ "add-files",		no_argument,		NULL, 'a'},
 	{ "tape",		required_argument,	NULL, 'T'},
 	{ "dry-run",		no_argument,		NULL, '0'},
+	{ "force",		no_argument,		NULL, 'f'},
 	{ NULL,			0,			NULL, 0 }
 };
 
 /* Command line option abbreviations */
-static const char option_string[] = "-c:t:i:r:p:P:d:D:s:m:hHnVvaT:";
+static const char option_string[] = "-c:t:i:r:p:P:d:D:M:s:m:hHnVvaT:f";
 
 struct command_line {
 	char* data[SCAN_KEYWORD_NUM];
@@ -61,6 +68,7 @@ struct command_line {
 	int verbose;
 	int add_files;
 	int dry_run;
+	int force;
 	enum scan_section_type type;
 };
 
@@ -106,6 +114,16 @@ get_command_line(int argc, char* argv[], struct command_line* line)
 			rc = store_option(&cmdline, scan_keyword_dumptofs,
 					  optarg);
 			break;
+		case 'M':
+			is_keyword = 1;
+			rc = store_option(&cmdline, scan_keyword_mvdump,
+					  optarg);
+#ifndef __s390x__
+			error_reason("Option '%s' not supported on 31-bit",
+				     scan_keyword_name(scan_keyword_mvdump));
+			rc = -1;
+#endif
+			break;
 		case 'i':
 			is_keyword = 1;
 			rc = store_option(&cmdline, scan_keyword_image,
@@ -135,6 +153,32 @@ get_command_line(int argc, char* argv[], struct command_line* line)
 			rc = store_option(&cmdline, scan_keyword_target,
 					  optarg);
 			break;
+		case 0xaa:
+			is_keyword = 1;
+			rc = store_option(&cmdline, scan_keyword_targetbase,
+					  optarg);
+			break;
+		case 0xab:
+			is_keyword = 1;
+			rc = store_option(&cmdline, scan_keyword_targettype,
+					  optarg);
+			break;
+		case 0xac:
+			is_keyword = 1;
+			rc = store_option(&cmdline, scan_keyword_targetgeometry,
+					  optarg);
+			break;
+		case 0xad:
+			is_keyword = 1;
+			rc = store_option(&cmdline,
+					  scan_keyword_targetblocksize,
+					  optarg);
+			break;
+		case 0xae:
+			is_keyword = 1;
+			rc = store_option(&cmdline, scan_keyword_targetoffset,
+					  optarg);
+			break;
 		case 'T':
 			is_keyword = 1;
 			rc = store_option(&cmdline, scan_keyword_tape,
@@ -174,6 +218,9 @@ get_command_line(int argc, char* argv[], struct command_line* line)
 		case '0':
 			cmdline.dry_run = 1;
 			break;
+		case 'f':
+			cmdline.force = 1;
+			break;
 		case 1:
 			/* Non-option is interpreted as section name */
 			if (cmdline.section != NULL) {
@@ -235,17 +282,27 @@ get_command_line(int argc, char* argv[], struct command_line* line)
 		if (rc) {
 			if (cmdline.type == section_invalid) {
 				error_reason("Need one of options 'image', "
-					     "'segment','dumpto', 'dumptofs' "
-					     "or 'menu'");
+					     "'segment','dumpto', 'dumptofs', "
+					     "'mvdump' or 'menu'");
 			}
 			return rc;
 		}
+		rc = scan_check_target_data(cmdline.data, NULL);
+		if (rc)
+			return rc;
 	}
 	*line = cmdline;
 	return 0;
 }
 
 
+static void
+free_target_data(struct job_target_data* data)
+{
+	if (data->targetbase != NULL)
+		free(data->targetbase);
+}
+
 static void
 free_ipl_data(struct job_ipl_data* data)
 {
@@ -324,11 +381,25 @@ free_dump_data(struct job_dump_data* data)
 }
 
 
+void
+free_mvdump_data(struct job_mvdump_data* data)
+{
+	int i;
+
+	if (data->device_list != NULL)
+		free(data->device_list);
+	for (i = 0; i < data->device_count; i++)
+		if (data->device[i] != NULL)
+			free(data->device[i]);
+}
+
+
 void
 job_free(struct job_data* job)
 {
-	if (job->bootmap_dir != NULL)
-		free(job->bootmap_dir);
+	if (job->target.bootmap_dir != NULL)
+		free(job->target.bootmap_dir);
+	free_target_data(&job->target);
 	if (job->name != NULL)
 		free(job->name);
 	switch (job->id) {
@@ -344,6 +415,9 @@ job_free(struct job_data* job)
 	case job_dump_partition:
 		free_dump_data(&job->data.dump);
 		break;
+	case job_mvdump:
+		free_mvdump_data(&job->data.mvdump);
+		break;
 	case job_dump_fs:
 		free_dump_fs_data(&job->data.dump_fs);
 		break;
@@ -481,6 +555,7 @@ check_job_menu_data(struct job_menu_data* menu)
 		case job_print_version:
 		case job_segment:
 		case job_dump_partition:
+		case job_mvdump:
 		case job_menu:
 		case job_ipl_tape:
 			break;
@@ -534,6 +609,67 @@ check_job_ipl_tape_data(struct job_ipl_tape_data *ipl, char* name)
 	return 0;
 }
 
+static int
+check_job_mvdump_data(struct job_mvdump_data* dump, char* name)
+{
+	int rc;
+	size_t size, i, j, k;
+	char* buffer;
+
+	rc = misc_read_file(dump->device_list, &buffer, &size, 0);
+	if (rc) {
+		if (name == NULL) {
+			error_text("Dump target list '%s'", dump->device_list);
+		} else {
+			error_text("Dump target list '%s' in section '%s'",
+				   dump->device_list, name);
+		}
+		return rc;
+	}
+	if (size == 0) {
+		error_text("Dump target list '%s' is empty.",
+			   dump->device_list);
+		free(buffer);
+		return -1;
+	}
+	for (i = 0, j = 0, k = 0 ; i < size; i++) {
+		if (buffer[i] != '\n')
+			buffer[j++] = buffer[i];
+		else if (j > 0) {
+			if (k == MAX_DUMP_VOLUMES) {
+				error_text("Dump target list '%s' contains "
+					   "more than %d entries.",
+					   dump->device_list, k);
+				free(buffer);
+				return -1;
+			}
+			buffer[j] = 0;
+			dump->device[k] = misc_strdup(buffer);
+			if (dump->device[k] == NULL)
+				return -1;
+			k++;
+			j = 0;
+		}
+	}
+	free(buffer);
+	dump->device_count = k;
+	for (i = 0; i < (size_t) dump->device_count; i++) {
+		rc = misc_check_writable_device(dump->device[i], 1, 0);
+		if (rc) {
+			error_text("Dump target '%s'", dump->device[i]);
+			return rc;
+		}
+		for (j = 0; j < i; j++) {
+			if (!strcmp(dump->device[i], dump->device[j])) {
+				error_text("Dump target list '%s' contains "
+					   "duplicate entry '%s'.",
+					   dump->device_list, dump->device[i]);
+				return -1;
+			}
+		}
+	}
+	return 0;
+}
 
 static int
 check_job_data(struct job_data* job)
@@ -541,16 +677,16 @@ check_job_data(struct job_data* job)
 	int rc;
 
 	/* Check for missing information */
-	if (job->bootmap_dir != NULL) {
-		rc = misc_check_writable_directory(job->bootmap_dir);
+	if (job->target.bootmap_dir != NULL) {
+		rc = misc_check_writable_directory(job->target.bootmap_dir);
 		if (rc) {
 			if (job->name == NULL) {
 				error_text("Target directory '%s'",
-					   job->bootmap_dir);
+					   job->target.bootmap_dir);
 			} else {
 				error_text("Target directory '%s' in section "
 					   "'%s'",
-					   job->bootmap_dir, job->name);
+					   job->target.bootmap_dir, job->name);
 			}
 			return rc;
 		}
@@ -571,6 +707,8 @@ check_job_data(struct job_data* job)
 		return check_job_dump_fs_data(&job->data.dump_fs, job->name);
 	case job_ipl_tape:
 		return check_job_ipl_tape_data(&job->data.ipl_tape, job->name);
+	case job_mvdump:
+		return check_job_mvdump_data(&job->data.mvdump, job->name);
 	}
 	return 0;
 }
@@ -726,6 +864,28 @@ get_parmline(char* filename, char* line, char** parmline, address_t* address,
 
 #define	MEGABYTE_MASK	(1024LL * 1024LL - 1LL)
 
+int
+type_from_target(char *target, disk_type_t *type)
+{
+	switch (scan_get_target_type(target)) {
+	case target_type_scsi:
+		*type = disk_type_scsi;
+		return 0;
+	case target_type_fba:
+		*type = disk_type_fba;
+		return 0;
+	case target_type_ldl:
+		*type = disk_type_eckd_classic;
+		return 0;
+	case target_type_cdl:
+		*type = disk_type_eckd_compatible;
+		return 0;
+	default:
+		return -1;
+	}
+}
+
+
 static int
 get_job_from_section_data(char* data[], struct job_data* job, char* section)
 {
@@ -736,10 +896,39 @@ get_job_from_section_data(char* data[], struct job_data* job, char* section)
 		/* IPL job */
 		job->id = job_ipl;
 		/* Fill in name of bootmap directory */
-		job->bootmap_dir = misc_strdup(data[(int) scan_keyword_target]);
-		if (job->bootmap_dir == NULL)
+		job->target.bootmap_dir =
+			misc_strdup(data[(int) scan_keyword_target]);
+		if (job->target.bootmap_dir == NULL)
 			return -1;
+		/* Fill in target */
+		if (data[(int) scan_keyword_targetbase] != NULL) {
+			job->target.targetbase =
+				misc_strdup(data[(int)
+				scan_keyword_targetbase]);
+			if (job->target.targetbase == NULL)
+				return -1;
+		}
+		if (data[(int) scan_keyword_targettype] != NULL) {
+			if (type_from_target(
+				data[(int) scan_keyword_targettype],
+				&job->target.targettype))
+				return -1;
+		}
+		if (data[(int) scan_keyword_targetgeometry] != NULL) {
+			job->target.targetcylinders =
+				atoi(strtok(data[(int)
+				scan_keyword_targetgeometry], ","));
+			job->target.targetheads = atoi(strtok(NULL, ","));
+			job->target.targetsectors = atoi(strtok(NULL, ","));
+		}
+		if (data[(int) scan_keyword_targetblocksize] != NULL)
+			job->target.targetblocksize =
+				atoi(data[(int) scan_keyword_targetblocksize]);
+		if (data[(int) scan_keyword_targetoffset] != NULL)
+			job->target.targetoffset =
+				atol(data[(int) scan_keyword_targetoffset]);
 		/* Fill in name and address of image file */
+
 		job->data.ipl.image = misc_strdup(
 					data[(int) scan_keyword_image]);
 		if (job->data.ipl.image == NULL)
@@ -807,8 +996,9 @@ get_job_from_section_data(char* data[], struct job_data* job, char* section)
 		/* SEGMENT LOAD job */
 		job->id = job_segment;
 		/* Fill in name of bootmap directory */
-		job->bootmap_dir = misc_strdup(data[(int) scan_keyword_target]);
-		if (job->bootmap_dir == NULL)
+		job->target.bootmap_dir =
+			misc_strdup(data[(int) scan_keyword_target]);
+		if (job->target.bootmap_dir == NULL)
 			return -1;
 		/* Fill in segment filename */
 		job->data.segment.segment =
@@ -844,8 +1034,9 @@ get_job_from_section_data(char* data[], struct job_data* job, char* section)
 		/* DUMP TO FILESYSTEM job */
 		job->id = job_dump_fs;
 		/* Fill in name of bootmap directory */
-		job->bootmap_dir = misc_strdup(data[(int) scan_keyword_target]);
-		if (job->bootmap_dir == NULL)
+		job->target.bootmap_dir =
+			misc_strdup(data[(int) scan_keyword_target]);
+		if (job->target.bootmap_dir == NULL)
 			return -1;
 		/* Fill in partition name */
 		job->data.dump_fs.partition =
@@ -872,6 +1063,27 @@ get_job_from_section_data(char* data[], struct job_data* job, char* section)
 		if (rc)
 			return rc;
 		break;
+	case section_mvdump:
+		/* DUMP TO MULTI-VOLUME job */
+		job->id = job_mvdump;
+		/* Fill in name of file containing the multiple partitions */
+		job->data.mvdump.device_list = misc_strdup(
+					       data[(int) scan_keyword_mvdump]);
+		if (job->data.mvdump.device_list == NULL)
+			return -1;
+		/* Check for mem size specification */
+		if (extract_memsize(job->data.mvdump.device_list,
+				    &job->data.mvdump.mem) == 0) {
+			/* Ensure megabyte alignment of size */
+			job->data.mvdump.mem =
+				(job->data.mvdump.mem + MEGABYTE_MASK) &
+					~MEGABYTE_MASK;
+			printf("Setting dump size limit to %lldMB\n",
+			       (unsigned long long) job->data.mvdump.mem /
+			       		(1024LL * 1024LL));
+		} else
+			job->data.mvdump.mem = -1LL;
+		break;
 	default:
 		/* Should not happen */
 		job->id = job_print_usage;
@@ -929,11 +1141,43 @@ get_menu_job(struct scan_token* scan, char* menu, struct job_data* job)
 					  atol(scan[i].content.keyword.value);
 					break;
 				case scan_keyword_target:
-					job->bootmap_dir = misc_strdup(
+					job->target.bootmap_dir = misc_strdup(
 						scan[i].content.keyword.value);
-					if (job->bootmap_dir == NULL)
+					if (job->target.bootmap_dir == NULL)
 						return -1;
 					break;
+				case scan_keyword_targetbase:
+					job->target.targetbase = misc_strdup(
+						scan[i].content.keyword.value);
+					if (job->target.targetbase == NULL)
+						return -1;
+					break;
+				case scan_keyword_targettype:
+					if (type_from_target(
+						scan[i].content.keyword.value,
+						&job->target.targettype))
+						return -1;
+					break;
+				case scan_keyword_targetgeometry:
+					job->target.targetcylinders =
+						atoi(strtok(
+						scan[i].content.keyword.value,
+						","));
+					job->target.targetheads =
+						atoi(strtok(NULL, ","));
+					job->target.targetsectors =
+						atoi(strtok(NULL, ","));
+					break;
+				case scan_keyword_targetblocksize:
+					job->target.targetblocksize =
+						atoi(
+						scan[i].content.keyword.value);
+					break;
+				case scan_keyword_targetoffset:
+					job->target.targetoffset =
+						atol(
+						scan[i].content.keyword.value);
+					break;
 				default:
 					/* Should not happen */
 					break;
@@ -1224,6 +1468,7 @@ job_get(int argc, char* argv[], struct job_data** data)
 	job->noninteractive = cmdline.noninteractive;
 	job->verbose = cmdline.verbose;
 	job->add_files = cmdline.add_files;
+	job->data.mvdump.force = cmdline.force;
 	job->dry_run = cmdline.dry_run;
 	/* Get job data from user input */
 	if (cmdline.help) {
