@@ -2,7 +2,7 @@
  * s390-tools/zipl/src/scan.c
  *   Scanner for zipl.conf configuration files
  *
- * Copyright IBM Corp. 2001, 2006.
+ * Copyright IBM Corp. 2001, 2009.
  *
  * Author(s): Carsten Otte <cotte@de.ibm.com>
  *            Peter Oberparleiter <Peter.Oberparleiter@de.ibm.com>
@@ -28,22 +28,36 @@
 
 /* Determines which keyword may be present in which section */
 enum scan_key_state scan_key_table[SCAN_SECTION_NUM][SCAN_KEYWORD_NUM] = {
-/*	 defa dump dump imag para parm ramd segm targ prom time defa tape
- *	 ult  to   tofs e    mete file isk  ent  et   pt   out  ultm
+/*	 defa dump dump imag para parm ramd segm targ prom time defa tape mv
+ *	 ult  to   tofs e    mete file isk  ent  et   pt   out  ultm      dump
  *			     rs                                 enu
+ *
+ *       targ targ targ targ targ
+ *       etba etty etge etbl etof
+ *       se   pe   omet ocks fset
+ *                 ry   ize
  */
 /* defaultboot	*/
-	{opt, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, opt, inv},
+	{opt, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, opt, inv, inv,
+	 inv, inv, inv, inv, inv},
 /* ipl		*/
-	{inv, inv, inv, req, opt, opt, opt, inv, req, inv, inv, inv, inv},
+	{inv, inv, inv, req, opt, opt, opt, inv, req, inv, inv, inv, inv, inv,
+	 opt, opt, opt, opt, opt},
 /* segment load */
-	{inv, inv, inv, inv, inv, inv, inv, req, req, inv, inv, inv, inv},
+	{inv, inv, inv, inv, inv, inv, inv, req, req, inv, inv, inv, inv, inv,
+	 inv, inv, inv, inv, inv},
 /* part dump	*/
-	{inv, req, inv, inv, inv, inv, inv, inv, opt, inv, inv, inv, inv},
+	{inv, req, inv, inv, inv, inv, inv, inv, opt, inv, inv, inv, inv, inv,
+	 inv, inv, inv, inv, inv},
 /* fs dump	*/
-	{inv, inv, req, inv, opt, opt, inv, inv, req, inv, inv, inv, inv},
+	{inv, inv, req, inv, opt, opt, inv, inv, req, inv, inv, inv, inv, inv,
+	 inv, inv, inv, inv, inv},
 /* ipl tape	*/
-	{inv, inv, inv, req, opt, opt, opt, inv, inv, inv, inv, inv, req}
+	{inv, inv, inv, req, opt, opt, opt, inv, inv, inv, inv, inv, req, inv,
+	 inv, inv, inv, inv, inv},
+/* multi volume dump	*/
+	{inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, req,
+	 inv, inv, inv, inv, inv}
 };
 
 /* Mapping of keyword IDs to strings */
@@ -56,17 +70,22 @@ static const struct {
 	{ "dumptofs", scan_keyword_dumptofs },
 	{ "dumpto", scan_keyword_dumpto },
 	{ "image", scan_keyword_image },
+	{ "mvdump", scan_keyword_mvdump },
 	{ "parameters", scan_keyword_parameters },
 	{ "parmfile", scan_keyword_parmfile },
 	{ "ramdisk", scan_keyword_ramdisk },
 	{ "segment", scan_keyword_segment },
+	{ "targetbase", scan_keyword_targetbase},
+	{ "targettype", scan_keyword_targettype},
+	{ "targetgeometry", scan_keyword_targetgeometry},
+	{ "targetblocksize", scan_keyword_targetblocksize},
+	{ "targetoffset", scan_keyword_targetoffset},
 	{ "target", scan_keyword_target},
 	{ "prompt", scan_keyword_prompt},
 	{ "timeout", scan_keyword_timeout},
 	{ "tape", scan_keyword_tape}
 };
 
-
 /* Retrieve name of keyword identified by ID. */
 char *
 scan_keyword_name(enum scan_keyword_id id)
@@ -599,10 +618,25 @@ scan_get_section_type(char* keyword[])
 		return section_dump;
 	else if (keyword[(int) scan_keyword_dumptofs] != NULL)
 		return section_dumpfs;
+	else if (keyword[(int) scan_keyword_mvdump] != NULL)
+		return section_mvdump;
 	else
 		return section_invalid;
 }
 
+enum scan_target_type
+scan_get_target_type(char *type)
+{
+	if (strcasecmp(type, "SCSI") == 0)
+		return target_type_scsi;
+	else if (strcasecmp(type, "FBA") == 0)
+		return target_type_fba;
+	else if (strcasecmp(type, "LDL") == 0)
+		return target_type_ldl;
+	else if (strcasecmp(type, "CDL") == 0)
+		return target_type_cdl;
+	return target_type_invalid;
+}
 
 #define MAX(a,b)	((a)>(b)?(a):(b))
 
@@ -635,9 +669,16 @@ scan_check_section_data(char* keyword[], int* line, char* name,
 		} else if (keyword[(int) scan_keyword_dumptofs]) {
 			*type = section_dumpfs;
 			main_keyword = scan_keyword_name(scan_keyword_dumptofs);
-		} else
-			/* Incomplete section data */
+		} else if (keyword[(int) scan_keyword_mvdump]) {
+			*type = section_mvdump;
+			main_keyword = scan_keyword_name(scan_keyword_mvdump);
+		} else {
+			error_reason("Line %d: section '%s' must contain "
+				     "either one of keywords 'image', "
+				     "'segment', 'dumpto', 'dumptofs', "
+				     "'mvdump' or 'tape'", section_line, name);
 			return -1;
+		}
 	}
 	/* Check keywords */
 	for (i=0; i < SCAN_KEYWORD_NUM; i++) {
@@ -726,6 +767,163 @@ scan_check_section_data(char* keyword[], int* line, char* name,
 }
 
 
+static int
+check_blocksize(int size)
+{
+	switch (size) {
+	case 512:
+	case 1024:
+	case 2048:
+	case 4096:
+		return 0;
+	}
+	return -1;
+}
+
+
+int
+scan_check_target_data(char* keyword[], int* line)
+{
+	int cylinders, heads, sectors;
+	char dummy;
+	int number;
+	enum scan_keyword_id errid;
+
+	if (keyword[(int) scan_keyword_targetbase] == 0) {
+		if (keyword[(int) scan_keyword_targettype] != 0)
+			errid = scan_keyword_targettype;
+		else if ((keyword[(int) scan_keyword_targetgeometry] != 0))
+			errid = scan_keyword_targetgeometry;
+		else if ((keyword[(int) scan_keyword_targetblocksize] != 0))
+			errid = scan_keyword_targetblocksize;
+		else if ((keyword[(int) scan_keyword_targetoffset] != 0))
+			errid = scan_keyword_targetoffset;
+		else
+			return 0;
+		if (line != NULL)
+			error_reason("Line %d: keyword 'targetbase' required "
+				"when specifying '%s'",
+				line[(int) errid], scan_keyword_name(errid));
+		else
+			error_reason("Option 'targetbase' required when "
+				"specifying '%s'",
+				scan_keyword_name(errid));
+		return -1;
+	}
+	if (keyword[(int) scan_keyword_targettype] == 0) {
+		if (line != NULL)
+			error_reason("Line %d: keyword 'targettype' "
+				"required when specifying 'targetbase'",
+				line[(int) scan_keyword_targetbase]);
+		else
+			error_reason("Option 'targettype' required "
+				     "when specifying 'targetbase'");
+		return -1;
+	}
+	switch (scan_get_target_type(keyword[(int) scan_keyword_targettype])) {
+	case target_type_cdl:
+	case target_type_ldl:
+		if ((keyword[(int) scan_keyword_targetgeometry] != 0))
+			break;
+		if (line != NULL)
+			error_reason("Line %d: keyword 'targetgeometry' "
+				"required when specifying 'targettype' %s",
+				line[(int) scan_keyword_targettype],
+				keyword[(int) scan_keyword_targettype]);
+		else
+			error_reason("Option 'targetgeometry' required when "
+				"specifying 'targettype' %s",
+				keyword[(int) scan_keyword_targettype]);
+		return -1;
+	case target_type_scsi:
+	case target_type_fba:
+		if ((keyword[(int) scan_keyword_targetgeometry] == 0))
+			break;
+		if (line != NULL)
+			error_reason("Line %d: keyword "
+				"'targetgeometry' not allowed for "
+				"'targettype' %s",
+				line[(int) scan_keyword_targetgeometry],
+				keyword[(int) scan_keyword_targettype]);
+		else
+			error_reason("Keyword 'targetgeometry' not "
+				"allowed for 'targettype' %s",
+				keyword[(int) scan_keyword_targettype]);
+		return -1;
+	case target_type_invalid:
+		if (line != NULL)
+			error_reason("Line %d: Unrecognized 'targettype' value "
+				"'%s'",
+				line[(int) scan_keyword_targettype],
+				keyword[(int) scan_keyword_targettype]);
+		else
+			error_reason("Unrecognized 'targettype' value '%s'",
+				keyword[(int) scan_keyword_targettype]);
+		return -1;
+	}
+	if (keyword[(int) scan_keyword_targetgeometry] != 0) {
+		if ((sscanf(keyword[(int) scan_keyword_targetgeometry],
+		    "%d,%d,%d %c", &cylinders, &heads, &sectors, &dummy)
+		    != 3) || (cylinders <= 0) || (heads <= 0) ||
+		    (sectors <= 0)) {
+			if (line != NULL)
+				error_reason("Line %d: Invalid target geometry "
+					"'%s'", line[
+					(int) scan_keyword_targetgeometry],
+					keyword[
+					(int) scan_keyword_targetgeometry]);
+			else
+				error_reason("Invalid target geometry '%s'",
+					keyword[
+					(int) scan_keyword_targetgeometry]);
+			return -1;
+		}
+	}
+	if (keyword[(int) scan_keyword_targetblocksize] == 0) {
+		if (line != NULL)
+			error_reason("Line %d: Keyword 'targetblocksize' "
+				"required when specifying 'targetbase'",
+				line[(int) scan_keyword_targetbase]);
+		else
+			error_reason("Option 'targetblocksize' required when "
+				"specifying 'targetbase'");
+		return -1;
+	}
+	if ((sscanf(keyword[(int) scan_keyword_targetblocksize], "%d %c",
+	    &number, &dummy) != 1) || check_blocksize(number)) {
+		if (line != NULL)
+			error_reason("Line %d: Invalid target blocksize '%s'",
+				line[(int) scan_keyword_targetblocksize],
+				keyword[(int) scan_keyword_targetblocksize]);
+		else
+			error_reason("Invalid target blocksize '%s'",
+				keyword[(int) scan_keyword_targetblocksize]);
+		return -1;
+	}
+	if (keyword[(int) scan_keyword_targetoffset] == 0) {
+		if (line != NULL)
+			error_reason("Line %d: Keyword 'targetoffset' "
+				"required when specifying 'targetbase'",
+				line[(int) scan_keyword_targetbase]);
+		else
+			error_reason("Option 'targetoffset' required when "
+				"specifying 'targetbase'");
+		return -1;
+	}
+	if (sscanf(keyword[(int) scan_keyword_targetoffset], "%d %c",
+	    &number, &dummy) != 1) {
+		if (line != NULL)
+			error_reason("Line %d: Invalid target offset '%s'",
+				line[(int) scan_keyword_targetoffset],
+				keyword[(int) scan_keyword_targetoffset]);
+		else
+			error_reason("Invalid target offset '%s'",
+				keyword[(int) scan_keyword_targetoffset]);
+		return -1;
+	}
+	return 0;
+}
+
 /* Check section at INDEX for compliance with config file rules. Upon success,
  * return zero and advance INDEX to point to the end of the section. Return
  * non-zero otherwise. */
@@ -794,13 +992,10 @@ check_section(struct scan_token* scan, int* index)
 		}
 	}
 	rc = scan_check_section_data(keyword, keyword_line, name, line, &type);
-	/* Check for missing keyword */
-	if (type == section_invalid) {
-		error_reason("Line %d: section '%s' must contain either one "
-			     "of keywords 'image', 'segment', 'dumpto', "
-			     "'dumptofs' or 'tape'", line, name);
-		return -1;
-	}
+	if (rc)
+		return rc;
+	/* Check target data */
+	rc = scan_check_target_data(keyword, keyword_line);
 	if (rc)
 		return rc;
 	/* Advance index to end of section */
@@ -832,6 +1027,8 @@ find_num_assignment(struct scan_token* scan, int num, int offset)
 static int
 check_menu(struct scan_token* scan, int* index)
 {
+	char* keyword[SCAN_KEYWORD_NUM];
+	int keyword_line[SCAN_KEYWORD_NUM];
 	enum scan_keyword_id key_id;
 	char* name;
 	char* str;
@@ -844,6 +1041,7 @@ check_menu(struct scan_token* scan, int* index)
 	int is_default;
 	int is_prompt;
 	int is_timeout;
+	int rc;
 
 	i = *index;
 	name = scan[i].content.menu.name;
@@ -878,6 +1076,24 @@ check_menu(struct scan_token* scan, int* index)
 				}
 				is_target = 1;
 				break;
+			case scan_keyword_targetbase:
+			case scan_keyword_targettype:
+			case scan_keyword_targetgeometry:
+			case scan_keyword_targetblocksize:
+			case scan_keyword_targetoffset:
+				key_id = scan[i].content.keyword.keyword;
+				keyword_line[key_id] = scan[i].line;
+				/* Rule 5 */
+				if (keyword[(int) key_id] != NULL) {
+					error_reason("Line %d: keyword '%s' "
+						"already specified",
+						scan[i].line,
+						scan_keyword_name(key_id));
+					return -1;
+				}
+				keyword[(int) key_id] =
+					scan[i].content.keyword.value;
+				break;
 			case scan_keyword_default:
 				if (is_default) {
 					error_reason("Line %d: keyword '%s' "
@@ -1036,6 +1252,10 @@ check_menu(struct scan_token* scan, int* index)
 			     name);
 		return -1;
 	}
+	/* Check target data */
+	rc = scan_check_target_data(keyword, keyword_line);
+	if (rc)
+		return rc;
 	/* Advance index to end of menu section */
 	*index = i - 1;
 	return 0;
